/*
 * XMMS meta-input plugin: A unix program interface for xmms input plugins
 *
 * Copyright (C) 2000 Mikael Bouillot <mikael dot bouillot at bigfoot dot com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>

#include "xmms/plugin.h"

InputPlugin *get_iplugin_info (void);
static void init (void);
static int is_our_file (char *filename);
static void *play_loop(void *arg);
static void play (char *filename);
static void stop (void);
static void pause (short p);
static int get_time (void);
static void get_song_info (char *filename, char **title_real, int *len_real);

InputPlugin meta_input_ip =
{
	NULL,
	NULL,
	"~Meta-input plugin", /* You REALLY want this at the end of the list, believe me */
	init,
	NULL,
	NULL,
	is_our_file,
	NULL,
	play,
	stop,
	pause,
	NULL,
	NULL,
	get_time,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	get_song_info,
	NULL,
	NULL
};

static pthread_t decode_thread;
static gboolean audio_error = FALSE;

static FILE *file;
int going, eof;
char data [2048];
int written;
int playtime;
int datapipe;
char *players [32];
int numPlayers;

InputPlugin *get_iplugin_info (void)
{
	return &meta_input_ip;
}

/*
 *  We look for all available players and store them in the
 * "players" global.
 */
static void init (void)
{
    char filename [256];
    DIR *dir;
    struct dirent *ent;
    struct stat statbuf;

	numPlayers = 0;
	
	dir = opendir (PLAYER_DIR);
	if (!dir)
		return;
	
	while ((ent = readdir (dir)) != NULL)
	{
		sprintf (filename, "%s/%s", PLAYER_DIR, ent->d_name);
		if (stat (filename, &statbuf) != -1 && S_ISREG (statbuf.st_mode))
		{
			players [numPlayers] = strdup (filename);
			numPlayers++;
		}
	}
	
	closedir (dir);
}

/*
 *  Which player knows how to play this file? Let's ask them.
 * Yes, this really ought to be cached!
 */
static int whose_file (char *filename)
{
	char command [256];
	int a;
	
	for (a = 0; a < numPlayers; a++)
	{
		sprintf (command, "%s isOurFile \"%s\"", players [a], filename);
		if (system (command) == 0)
			return a;
	}
	
	return -1;
}

/*
 *  It's our file if one player knows how to play it.
 */
static int is_our_file (char *filename)
{
	int a;
	
	if (whose_file (filename) != -1)
		return TRUE;
	else
		return FALSE;
}

/*
 *  The play loop. Executed in a separate thread, it just shovels
 * what comes from the data pipe to the vis and output plugins.
 * When the pipe is closed, we know the player died.
 */
static void *play_loop (void *arg)
{
	int a, nread;
	
	while (going)
	{
		if (!eof)
		{
			while (written < 2048)
			{
				nread = read (datapipe, data + written, 2048 - written);
				if (nread == 0 || nread == -1)
				{
					eof = 1;
					break;
				}
				written += nread;
			}

			meta_input_ip.add_vis_pcm (meta_input_ip.output->written_time(), FMT_S16_LE, 2, written, data);

			while (going && meta_input_ip.output->buffer_free() < written)
				xmms_usleep (10000);

			if (going)
				meta_input_ip.output->write_audio (data, written);
			
			written = 0;
		}
		else	/* eof */
		{
			xmms_usleep (10000);
		}
	}
	
	pthread_exit (NULL);
}

/*
 *  We are asked to play a file, so we check what player can
 * play it, then we create the data pipe, we fork the player
 * with his standard output plugged into our pipe and we launch
 * the playing thread.
 */
static void play (char *filename)
{
	FILE *file;
	char *name, *temp;
	unsigned long len;
	int rate;
	int filedes [2];
	int ret;
	int player;
	int a;
	char command [256];
	
	audio_error = FALSE;
	
	player = whose_file (filename);
	
	if (player != -1)
	{
		written = 0;
		going = 1;
		eof = 0;
		if (meta_input_ip.output->open_audio (FMT_S16_LE, 44100, 2) == 0)
		{
			audio_error = TRUE;
			return;
		}

		pipe (filedes);

		ret = fork();

		if (ret == -1)
		{
			audio_error = TRUE;
			return;
		}

		if (ret == 0)	// Child process
		{
			close (filedes [0]);
			close (0);
			dup2 (filedes [1], 1);
			close (filedes [1]);
			
			execl (players [player], "foo", "play", filename, NULL);
			
			exit (-1);
		}

		// Parent process

		close (filedes [1]);

		datapipe = filedes [0];

		temp = strrchr (filename, '/');
		if (!temp)
			temp = filename;
		else
			temp++;
		name = malloc (strlen (temp) + 1);
		strcpy (name, temp);
		if (strrchr (name, '.') != NULL)
			*strrchr (name, '.') = '\0';

		meta_input_ip.set_info (name, -1, 333000 /* bps */, 44100, 2 /* stereo */);
		free (name);

		pthread_create (&decode_thread, NULL, play_loop, NULL);
	}
}

/*
 *  We have to stop playing. If we *are* playing, we suicide
 * the playing thread, we close the pipe, and we wait for both
 * the playing thread and the player to die.
 */
static void stop (void)
{
	if (going)
	{
		going = 0;
		close (datapipe);
		pthread_join (decode_thread, NULL);
		meta_input_ip.output->close_audio();
		wait (NULL);
	}
}

static void pause (short p)
{
	meta_input_ip.output->pause (p);
}

static int get_time (void)
{
	if (audio_error)
		return -2;
	if (!going || (eof && !meta_input_ip.output->buffer_playing()))
		return -1;
	else
		return meta_input_ip.output->output_time();
}

/*
 *  Generic info about one of our files. This ought to be
 * asked to the concerned player, so it's just done this
 * way until the player interface is improved.
 */
static void get_song_info (char *filename, char **title_real, int *len_real)
{
	FILE *file;
	char *name, *temp;
	int playtime, fadetime;

	temp = strrchr (filename, '/');
	if (!temp)
		temp = filename;
	else
		temp++;
	name = malloc (strlen (temp) + 1);
	strcpy (name, temp);
	if (strrchr (name, '.') != NULL)
		*strrchr (name, '.') = '\0';
	(*title_real) = name;
	(*len_real) = -1;
}
